import yaml

cvmfs_path = "/cvmfs/ctao.dpps.test"
apptainer_options = []

def translate(cwl):
    if cwl["class"] != "CommandLineTool":
        return cwl

    if (requirement := cwl.get("requirements", {}).pop("DockerRequirement", None)) is not None:

        image = requirement["dockerPull"]
        cmd = ["apptainer", "run", *apptainer_options, f"{cvmfs_path}/{image}"] + cwl["baseCommand"]
        cwl["baseCommand"] = cmd

    return cwl


if __name__ == "__main__":
    with open("./test.cwl") as f:
        cwl = yaml.safe_load(f)

    print(yaml.dump(translate(cwl)))



"""Test job status."""

import pytest

from wms.tests.utils import wait_for_status

pytestmark = [
    pytest.mark.wms,
    pytest.mark.dirac_client,
]


# missing "Run a single-job workflow" UC ID
# @pytest.mark.verifies_usecase("DPPS-UC-110-????")
@pytest.mark.usefixtures("_init_dirac")
def test_simple_job(tmp_path):
    from DIRAC.Interfaces.API.Dirac import Dirac
    from DIRAC.Interfaces.API.Job import Job

    dirac = Dirac()

    job = Job()
    job.setExecutable("echo", arguments="Hello world")
    job.setName("testjob")
    job.setDestination("CTAO.CI.de")
    res = dirac.submitJob(job)
    assert res["OK"]
    job_id = res["Value"]

    # wait for job to succeed, will error in case of timeout or job failure
    result = wait_for_status(
        dirac,
        job_id=job_id,
        status="Done",
        error_on={"Failed"},
        timeout=300,
        job_output_dir=tmp_path,
    )
    print(result)


@pytest.mark.usefixtures("_init_dirac")
def test_cvmfs_available_on_ce(tmp_path):
    from DIRAC.Interfaces.API.Dirac import Dirac
    from DIRAC.Interfaces.API.Job import Job

    dirac = Dirac()

    job = Job()
    job.setExecutable("ls", "/cvmfs/ctao.dpps.test/")
    job.setExecutable("cat", "/cvmfs/ctao.dpps.test/new_repository")
    job.setName("cvmfs_job")
    job.setDestination("CTAO.CI.de")
    res = dirac.submitJob(job)
    assert res["OK"]
    job_id = res["Value"]

    # wait for job to succeed, will error in case of timeout or job failure
    result = wait_for_status(
        dirac,
        job_id=job_id,
        status="Done",
        error_on={"Failed"},
        timeout=300,
        job_output_dir=tmp_path,
    )
    print(result)


@pytest.mark.usefixtures("_init_dirac")
def test_cwl_job(tmp_path):
    from DIRAC.Interfaces.API.Dirac import Dirac
    from DIRAC.Interfaces.API.Job import Job

    dirac = Dirac()

    job = Job()
    cwl_workflow = "src/wms/tests/hello_world.cwl"
    cwl_input = "src/wms/tests/input.yml"
    arguments_str = f"{cwl_workflow} {cwl_input}"
    job.setExecutable("cwltool", arguments=arguments_str, logFile="Job_Log.txt")
    job.setInputSandbox([cwl_workflow, cwl_input])
    job.setOutputSandbox(["*Log.txt"])
    job.setName("test_cwl_job")
    job.setDestination("CTAO.CI.de")
    res = dirac.submitJob(job)
    assert res["OK"]
    job_id = res["Value"]

    # wait for job to succeed, will error in case of timeout or job failure
    result = wait_for_status(
        dirac,
        job_id=job_id,
        status="Done",
        error_on={"Failed"},
        timeout=300,
        job_output_dir=tmp_path,
    )
    print(result)



import DIRAC
#from DIRAC.Interfaces.API.Dirac import Dirac
from cwl_job import CWLJob

DIRAC.initialize()  # Initialize configuration

job = CWLJob()

job.set_cwl("hello_world.cwl")
job.set_cwl_inputs("input.yml")
job.setTag("production")
res = job.submit()

#res = dirac.submitJob(job)

#print (res)
print(res["Value"])



import xml.etree.ElementTree as ET
from pathlib import Path

import pytest

from CTADIRAC.Interfaces.API.CWLJob import CWLJob
from CTADIRAC.Interfaces.Utilities.CWL_utils import (
    LFN_DIRAC_PREFIX,
    LFN_PREFIX,
)

INPUT_DATA = ["/ctao/user/MC/prod.sim", "/ctao/user/MC/prod2.sim"]
INPUT_SANDBOX = ["/a/local/MC/simulation.py", "/path/to/MC.prod3.sim"]
CWL_INPUTS_EXAMPLE = f"""
local_script:
  - class: File
    path: {INPUT_SANDBOX[0]}
input_as_lfn:
  - class: File
    path: {LFN_PREFIX}{INPUT_DATA[0]}
input_as_lfn_2:
  - class: File
    path: {LFN_PREFIX}{INPUT_DATA[1]}
  - class: File
    path: {INPUT_SANDBOX[1]}
dataset: "dataset://path/to/data"
input_param: "a random param"
"""
OUTPUT_DATA = ["/ctao/user/MC/fit*.out", "/ctao/usr/MC/data.sim"]
OUTPUT_SANDBOX = ["/path/to/test*.out", "/path/to/data_2.sim", "*.txt"]
BASE_COMMAND = "python"
IMAGE = "harbor.cta-observatory.org/proxy_cache/library/python:3.12-slim"
CWL_WORKFLOW_EXAMPLE = f"""
cwlVersion: v1.2
class: CommandLineTool
doc: |
      Test using local input/output data treated as Dirac
      input/output sandboxes.

inputs:
  local_script:
    type: File
    inputBinding:
      position: 1
  input_as_lfn:
    type: File?
    inputBinding:
      position: 2
  input_as_lfn_2:
    type: File[]
    inputBinding:
      position: 3
  dataset:
    type: [File, string]
    inputBinding:
      position: 4
  input_param:
    type: string
    inputBinding:
      position: 5

outputs:
  output_as_sb:
    type: File[]?
    outputBinding:
      glob: ["{OUTPUT_SANDBOX[0]}"]
  output_as_lfn:
    type: File?
    label: "LFN wildcards"
    outputBinding:
      glob: "{LFN_PREFIX}{OUTPUT_DATA[0]}"
  output_as_lfn_2:
    type: File[]
    label: "LFN files list"
    outputBinding:
      glob:
        - {LFN_PREFIX}{OUTPUT_DATA[1]}
        - {OUTPUT_SANDBOX[1]}
  output_as_array:
    type:
      type: array
      items: File
    outputBinding:
      glob: "{OUTPUT_SANDBOX[2]}"

baseCommand: ["{BASE_COMMAND}"]

hints:
  DockerRequirement:
    dockerPull: {IMAGE}
"""

CWL_WORKFLOW_EXAMPLE_NO_INPUT = """
cwlVersion: v1.2
class: CommandLineTool
doc: |
      Test using local input/output data treated as Dirac
      input/output sandboxes.

inputs:
  local_script:
    type: string
    inputBinding:
      position: 1

outputs:
  output_as_sb:
    type: File[]?
    outputBinding:
      glob: ["/path/to/output_1.txt"]

baseCommand: ["base_command"]
"""
CWL_WORKFLOW_EXAMPLE_NO_INPUT_INPUTS = """
local_script: "test"
"""


@pytest.fixture
def mock_submit_job(mocker):
    return mocker.patch(
        "CTADIRAC.Interfaces.API.CWLJob.Dirac.submitJob",
        side_effect=lambda self: self._toXML(),
    )


@pytest.mark.parametrize(
    (
        "cwl_worflow",
        "cwl_inputs",
        "expected_input_data",
        "expected_input_sandbox",
        "expected_output_data",
        "expected_output_sandbox",
    ),
    [
        (
            CWL_WORKFLOW_EXAMPLE,
            CWL_INPUTS_EXAMPLE,
            [f"{LFN_DIRAC_PREFIX}{data}" for data in INPUT_DATA],
            INPUT_SANDBOX,
            [f"{LFN_DIRAC_PREFIX}{data}" for data in OUTPUT_DATA],
            OUTPUT_SANDBOX,
        ),
        (
            CWL_WORKFLOW_EXAMPLE_NO_INPUT,
            CWL_WORKFLOW_EXAMPLE_NO_INPUT_INPUTS,
            [],
            [],
            [],
            ["/path/to/output_1.txt"],
        ),
    ],
)
def test_cwl_job_submit(
    mock_submit_job,
    tmp_path,
    cwl_worflow,
    cwl_inputs,
    expected_input_data,
    expected_input_sandbox,
    expected_output_data,
    expected_output_sandbox,
):
    cwl_workflow_example = tmp_path / "cwl_workflow_example.cwl"
    cwl_workflow_example.write_text(cwl_worflow)
    cwl_inputs_example = tmp_path / "cwl_inputs_example.cwl"
    cwl_inputs_example.write_text(cwl_inputs)

    job = CWLJob(
        cwl_workflow=cwl_workflow_example,
        cwl_inputs=cwl_inputs_example,
        cvmfs_base_path=Path("/cvmfs/ctao.dpps.test/"),
        output_se="TEST_SE",
    )

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    def get_list(name):
        element = result_xml.find(f".//Parameter[@name='{name}']/value")
        if element is None or element.text is None:
            return []
        return element.text.split(";")

    input_data = get_list("InputData")
    assert input_data == expected_input_data

    output_sandbox = get_list("OutputSandbox")
    assert set(expected_output_sandbox).issubset(output_sandbox)

    input_sandbox = get_list("InputSandbox")
    assert set(expected_input_sandbox).issubset(input_sandbox)

    output_data = get_list("OutputData")
    assert output_data == expected_output_data

    if expected_output_data:
        output_se_parameter = result_xml.find(".//Parameter[@name='OutputSE']")
        assert output_se_parameter is not None
        output_se = output_se_parameter.find("value")
        assert output_se is not None
        assert output_se.text == "TEST_SE"

    if executable := result_xml.find(
        ".//StepInstance/Parameter[@name='executable']/value"
    ):
        executable = executable.text
        assert executable == "cwltool"


def test_datapipe_cwl_job(mock_submit_job):
    job = CWLJob(
        "tests/resources/cwl/process_dl0_dl1.cwl",
        "tests/resources/cwl/inputs_process_dl0_dl1.yaml",
        cvmfs_base_path=Path("/cvmfs/ctao.dpps.test/"),
    )

    # these are set via inputs
    assert len(job.input_data) == 1
    assert (
        job.input_data[0]
        == "LFN:/ctao/simpipe/prod6/gamma-diffuse/010xxx/gamma_cone10_run010000.simtel.zst"
    )

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    if input_data := result_xml.find(".//Parameter[@name='InputData']/value"):
        input_data = input_data.text.split(";")
        assert len(input_data) == 1
        assert input_data[0] == "gamma_cone10_run010000.simtel.zst"

    if input_sandbox := result_xml.find(".//Parameter[@name='InputSandbox']/value"):
        input_sandbox = input_sandbox.text.split(";")
        assert len(input_sandbox) == 1
        assert input_sandbox[0] == "process_config.yaml"

    if output_data := result_xml.find(".//Parameter[@name='OutputData']/value"):
        output_data = output_data.text.split(";")
        assert len(output_data) == 1
        assert output_data[0] == "LFN:/ctao/datapipe/test.dl1.h5"

    if output_sandbox := result_xml.find(".//Parameter[@name='OutputSandbox']/value"):
        output_sandbox = output_sandbox.text.split(";")
        assert len(job.output_sandbox) == 1
        assert job.output_sandbox[0] == "ctapipe-process_dl0_dl1.provenance.log"



#!/usr/bin/env python
"""
Submit a CWL workflow

Usage:
cta-submit-job job.cwl job.yml

"""

__RCSID__ = "$Id$"


import os
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Interfaces.API.Job import Job
from CTADIRAC.ProductionSystem.CWL.CWLWorkflowStep import WorkflowStep
from DIRAC.Interfaces.API.Dirac import Dirac


def submit_job(run_cwl_file, run_cfg_file=None):
    # Build WorkflowSteps
    step = WorkflowStep()
    step.load(run_cwl_file, run_cfg_file)
    step.get_command_line()

    # Build Job
    job = Job()
    job.setName("test_cwl_job")

    input_sandbox = [run_cwl_file]
    if run_cfg_file:
        input_sandbox.append(run_cfg_file)
    #print(input_sandbox)
    job.setInputSandbox(input_sandbox)

    job.setOutputSandbox(["*Log.txt"])

    print("DEBUG_LA command line")
    print(step.command_line)
    job.setExecutable(
        str(step.command_line.split(" ", 1)[0]),
        arguments=str(step.command_line.split(" ", 1)[1]),
        logFile="Pipeline_Log.txt",
    )

    job.setTag("production")
    #print(job.workflow.toXML())
    # Submit job
    dirac = Dirac()

    res = dirac.submitJob(job)
    return res


def main():
    Script.parseCommandLine()
    argss = Script.getPositionalArgs()

    run_cwl_file = argss[0]
    run_cfg_file = None
    if len(argss)==2:
        run_cfg_file = argss[1]
    #print(run_cfg_file)

    res = submit_job(
        run_cwl_file,
        run_cfg_file,
    )

    try:
        if not res["OK"]:
            DIRAC.gLogger.error(res["Message"])
            DIRAC.exit(-1)
        else:
            gLogger.notice("Submission Result:", res["Value"])
    except Exception:
        DIRAC.gLogger.exception()
        DIRAC.exit(-1)


if __name__ == "__main__":
    main()



"""Utilities to handle CWL."""

from pathlib import Path
from typing import Any
from copy import deepcopy

from cwl_utils.parser import OutputArraySchema, File
from cwl_utils.expression import do_eval

LFN_PREFIX = "lfn://"
LFN_DIRAC_PREFIX = "LFN:"
LOCAL_PREFIX = "file://"


def fill_defaults(cwl, inputs):
    """Fill in defaults into inputs.

    This is needed for evaluating expressions later on.

    Parameters
    ----------
    cwl: dict
        The CWL definition
    inputs: dict
        user provided inputs.

    Returns
    -------
    inputs: dict
        inputs with additional values filled from CWL defaults
    """
    inputs_new = deepcopy(inputs)

    if cwl.inputs is None:
        return inputs_new

    for inp in cwl.inputs:
        key = inp.id.rpartition("#")[2]
        if key not in inputs and inp.default is not None:
            inputs_new[key] = inp.default
    return inputs_new


def translate_cwl_workflow(
    cwl_obj,
    cwl_inputs,
    cvmfs_base_path: Path,
    apptainer_options: list[Any],
) -> dict[str, Any]:
    """Translate the CWL workflow description into Dirac compliant execution."""
    output_dict = {"CWLDesc": cwl_obj, "OutputSandbox": [], "OutputData": []}

    if cwl_obj.class_ != "CommandLineTool":
        return output_dict

    if cwl_obj.hints:
        cwl_obj = translate_docker_hints(cwl_obj, cvmfs_base_path, apptainer_options)

    return extract_and_translate_output_files(cwl_obj, cwl_inputs)


def translate_docker_hints(
    cwl_obj, cvmfs_base_path: Path, apptainer_options: list[Any]
):
    """Translate CWL DockerRequirement into Dirac compliant execution."""
    for index, hints in enumerate(cwl_obj.hints):
        if hints.class_ == "DockerRequirement":
            image = hints.dockerPull
            cmd = [
                "apptainer",
                "run",
                *apptainer_options,
                str(cvmfs_base_path / f"{image}"),
            ]
            if isinstance(cwl_obj.baseCommand, str):
                cmd += [cwl_obj.baseCommand]
            else:
                cmd += cwl_obj.baseCommand
            cwl_obj.baseCommand = cmd
            del cwl_obj.hints[index]
    return cwl_obj


def collect_outputs(cwl, inputs) -> list[str]:
    """Collect evaluated output filenames.

    Parameters
    ----------
    cwl: dict
        The CWL definition
    inputs: dict
        user provided inputs.

    Returns
    -------
    outputs: list[str]
        The output filenames of this workflow given
        the provided inputs and defaults.
    """
    outputs = []

    if cwl.outputs is None:
        return outputs

    for output in cwl.outputs:
        if glob := output.outputBinding.glob:
            result = do_eval(
                glob,
                inputs,
                outdir=None,
                requirements=[],
                tmpdir=None,
                resources={},
            )
            outputs.append(result)

    return outputs


def extract_and_translate_output_files(cwl_obj, inputs) -> dict[str, Any]:
    """Translate output files into a DIRAC compliant usage.

    Extract local outputs and lfns.
    Remove outputs path prefix.
    """
    output_lfns = []
    output_sandboxes = []

    inputs = fill_defaults(cwl_obj, inputs)

    for outputs in cwl_obj.outputs or []:
        if not verify_cwl_output_type(outputs.type_):
            continue

        glob = outputs.outputBinding.glob

        if isinstance(glob, str):
            glob = do_eval(
                glob,
                inputs,
                outdir=None,
                requirements=[],
                tmpdir=None,
                resources={},
            )
            if glob.startswith(LFN_PREFIX):
                output_lfns.append(glob.replace(LFN_PREFIX, LFN_DIRAC_PREFIX))
            else:
                output_sandboxes.append(glob)

        if isinstance(glob, list):
            for glob_item in glob:
                glob_item = do_eval(
                    glob_item,
                    inputs,
                    outdir=None,
                    requirements=[],
                    tmpdir=None,
                    resources={},
                )
                if glob_item.startswith(LFN_PREFIX):
                    output_lfns.append(glob_item.replace(LFN_PREFIX, LFN_DIRAC_PREFIX))
                else:
                    output_sandboxes.append(glob_item)

    return {
        "CWLDesc": cwl_obj,
        "OutputSandbox": output_sandboxes,
        "OutputData": output_lfns,
    }


def verify_cwl_output_type(output_type) -> bool:
    """Verify the cwl output type.

    True if the type is File or OutputArraySchema
    or a list of 'null' and File/OutputArraySchema
    else False.
    """
    if isinstance(output_type, list):
        for type_ in output_type:
            if type_ == "File" or isinstance(type_, OutputArraySchema):
                return True
    if output_type == "File" or isinstance(output_type, OutputArraySchema):
        return True
    return False


# TODO: how to deal with default values not present in the input file but in the cwl?
def extract_and_translate_input_files(cwl_inputs) -> dict[str, Any]:
    """Extract input files from CWL inputs, rewrite file prefix.

    If the file is a Sandbox, ensure there is no absolute path,
    and store it in the input sandbox list.
    If the file is a LFN, remove the lfn prefix and store it in the lfns list.
    """
    input_sandboxes = []
    input_lfns = []
    cwl_inputs = deepcopy(cwl_inputs)

    def handle_file(input_value):
        original = input_value.path
        input_value.path, lfn = translate_sandboxes_and_lfns(original)
        if lfn:
            input_lfns.append(original.replace(LFN_PREFIX, LFN_DIRAC_PREFIX))
        else:
            input_sandboxes.append(original.removeprefix("file://"))

    for key, input_value in cwl_inputs.items():
        if isinstance(input_value, list):
            for input_file in input_value:
                handle_file(input_file)

        elif isinstance(input_value, File):
            handle_file(input_value)

        # assume that an LFN in the inputs as string is an output filename
        elif isinstance(input_value, str) and input_value.startswith(LFN_PREFIX):
            cwl_inputs[key], _ = translate_sandboxes_and_lfns(input_value)

    return {
        "InputDesc": cwl_inputs,
        "InputSandbox": input_sandboxes,
        "InputData": input_lfns,
    }


def translate_sandboxes_and_lfns(file: File | str) -> tuple[str, bool]:
    """Extract local files as sandboxes and lfns as input data."""
    if isinstance(file, File):
        if not file.path:
            raise KeyError("File path is not defined.")
        path = file.path
    elif isinstance(file, str):
        path = file

    lfn = False
    if path.startswith(LFN_PREFIX):
        path = path.removeprefix(LFN_PREFIX)
        lfn = True

    path = path.removeprefix(LOCAL_PREFIX)
    return Path(path).name, lfn



from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Interfaces.API.Job import Job

class CWLJob(Job):
    """Base Job class for CWL jobs"""

    def __init__(self) -> None:
        Job.__init__(self)
        self.setOutputSandbox(["*Log.txt"])
        self.setName("cwl_job")
        self.cwl_workflow: str = ""
        self.cwl_inputs: str = ""
        self.apptainer_repository: str = ""

    def set_cwl(self, cwl_workflow: str) -> None:
        """Set CWL workflow path or URL."""
        self.cwl_workflow = cwl_workflow

    def set_cwl_inputs(self, cwl_inputs: str) -> None:
        """Set CWL input parameters path."""
        self.cwl_inputs = cwl_inputs

    def set_apptainer_repository(self, apptainer_repository: str ) -> None:
        """Set the Apptainer repository path."""
        self.apptainer_repository = apptainer_repository

    def submit(self) -> dict:
        """Submit the CWL job to DIRAC."""
        dirac = Dirac()
        self.setInputSandbox([self.cwl_workflow, self.cwl_inputs])
        self.setExecutable("cwltool", arguments=f"{self.cwl_workflow} {self.cwl_inputs}", logFile="Job_Log.txt")
        res = dirac.submitJob(self)
        return res